@@ -242,6 +242,9 @@ cogl_onscreen_egl_swap_region (CoglOnscreen  *onscreen,
                                              n_rectangles,
                                              rectangles) == EGL_FALSE)
     g_warning ("Error reported by eglSwapBuffersRegion");
+
+  /* Update latest sync object after buffer swap */
+  cogl_framebuffer_flush (framebuffer);
 }
 
 static void
@@ -342,6 +345,9 @@ cogl_onscreen_egl_swap_buffers_with_damage (CoglOnscreen  *onscreen,
     }
   else
     eglSwapBuffers (egl_renderer->edpy, priv->egl_surface);
+
+  /* Update latest sync object after buffer swap */
+  cogl_framebuffer_flush (framebuffer);
 }
 
 void
