@@ -8,6 +8,11 @@ import subprocess
 import getpass
 import argparse
 import logind_helpers
+import tempfile
+import select
+import socket
+import threading
+import configparser
 from collections import OrderedDict
 from dbusmock import DBusTestCase
 from dbus.mainloop.glib import DBusGMainLoop
@@ -15,6 +20,14 @@ from pathlib import Path
 from gi.repository import Gio
 
 
+class MultiOrderedDict(OrderedDict):
+    def __setitem__(self, key, value):
+        if isinstance(value, list) and key in self:
+            self[key].extend(value)
+        else:
+            super(OrderedDict, self).__setitem__(key, value)
+
+
 def escape_object_path(path):
     b = bytearray()
     b.extend(path.encode())
@@ -36,7 +49,7 @@ class MutterDBusRunner(DBusTestCase):
             return os.path.join(os.path.dirname(__file__), 'dbusmock-templates')
 
     @classmethod
-    def setUpClass(klass, enable_kvm, launch):
+    def setUpClass(klass, enable_kvm=False, launch=[], bind_sockets=False):
         klass.templates_dirs = [klass.__get_templates_dir()]
 
         klass.mocks = OrderedDict()
@@ -56,10 +69,15 @@ class MutterDBusRunner(DBusTestCase):
         klass.start_session_bus()
         klass.start_system_bus()
 
+        klass.sockets = []
+        klass.poll_thread = None
+        if bind_sockets:
+            klass.enable_pipewire_sockets()
+
         print('Launching required services...', file=sys.stderr)
         klass.service_processes = []
         for service in launch:
-            klass.service_processes += [subprocess.Popen(service)]
+            klass.launch_service([service])
 
         print('Starting mocked services...', file=sys.stderr)
         (klass.mocks_manager, klass.mock_obj) = klass.start_from_local_template(
@@ -88,7 +106,12 @@ class MutterDBusRunner(DBusTestCase):
             mock_server.terminate()
             mock_server.wait()
 
+        print('Closing PipeWire socket...', file=sys.stderr)
+        klass.disable_pipewire_sockets()
+
+        print('Terminating services...', file=sys.stderr)
         for process in klass.service_processes:
+            print('  - Terminating {}'.format(' '.join(process.args)), file=sys.stderr)
             process.terminate()
             process.wait()
 
@@ -258,47 +281,185 @@ ret = logind_helpers.open_file_direct(major, minor)
                 return template_path
         raise FileNotFoundError(f'Couldnt find a {template_name} template')
 
+    @classmethod
+    def launch_service(klass, args, env=None, pass_fds=()):
+        print('  - Launching {}'.format(' '.join(args)), file=sys.stderr)
+        klass.service_processes += [subprocess.Popen(args, env=env, pass_fds=pass_fds)]
 
-def wrap_call(args, wrapper):
+    @classmethod
+    def poll_pipewire_sockets_in_thread(klass, sockets):
+        poller = select.poll()
+        for socket in sockets:
+            poller.register(socket.fileno(), select.POLLIN | select.POLLHUP)
+
+        should_spawn = False
+        should_poll = True
+        while should_poll:
+            results = poller.poll()
+            for result in results:
+                if result[1] == select.POLLIN:
+                    should_spawn = True
+                    should_poll = False
+                else:
+                    should_poll = False
+
+        if not should_spawn:
+            return
+
+        print("Noticed activity on a PipeWire socket, launching services...", file=sys.stderr);
+
+        pipewire_env = os.environ
+        pipewire_env['LISTEN_FDS'] = f'{len(sockets)}'
+        pipewire_fds = {}
+        subprocess_fd = 3
+        for sock in sockets:
+            pipewire_fds[subprocess_fd] = sock.fileno()
+            subprocess_fd += 1
+
+        socket_launch = os.path.join(os.path.dirname(__file__), 'socket-launch.sh')
+        klass.launch_service([socket_launch, 'pipewire'],
+                             env=pipewire_env,
+                             pass_fds=pipewire_fds)
+        klass.launch_service(['wireplumber'])
+
+    @classmethod
+    def get_pipewire_socket_names(klass):
+        pipewire_socket_unit = '/usr/lib/systemd/user/pipewire.socket'
+
+        config = configparser.ConfigParser(strict=False,
+                                           empty_lines_in_values=False,
+                                           dict_type=MultiOrderedDict,
+                                           interpolation=None)
+        res = config.read([pipewire_socket_unit])
+
+        runtime_dir = os.environ['XDG_RUNTIME_DIR']
+        return [socket_name.replace('%t', runtime_dir)
+                for socket_name in config.get('Socket', 'ListenStream')]
+
+    @classmethod
+    def enable_pipewire_sockets(klass):
+        runtime_dir = os.environ['XDG_RUNTIME_DIR']
+
+        sockets = []
+        for socket_name in klass.get_pipewire_socket_names():
+            sock = socket.socket(socket.AF_UNIX)
+            print("Binding {} for socket activation".format(socket_name), file=sys.stderr)
+            sock.bind(socket_name)
+            sock.listen()
+            sockets.append(sock)
+
+        poll_closure = lambda: klass.poll_pipewire_sockets_in_thread(sockets)
+
+        klass.poll_thread = threading.Thread(target=poll_closure)
+        klass.poll_thread.start()
+
+        klass.sockets = sockets
+
+    @classmethod
+    def disable_pipewire_sockets(klass):
+        for sock in klass.sockets:
+            sock.shutdown(socket.SHUT_RDWR)
+            sock.close()
+
+        if klass.poll_thread:
+            klass.poll_thread.join()
+
+
+def wrap_call(args, wrapper, extra_env):
     env = {}
     env.update(os.environ)
     env['NO_AT_BRIDGE'] = '1'
     env['GTK_A11Y'] = 'none'
     env['GSETTINGS_BACKEND'] = 'memory'
+    env['XDG_CURRENT_DESKTOP'] = ''
     env['META_DBUS_RUNNER_ACTIVE'] = '1'
 
+    if extra_env:
+        env |= extra_env
+
     if wrapper == 'gdb':
         args = ['gdb', '-ex', 'r', '-ex', 'bt full', '--args'] + args
+    elif wrapper == 'rr':
+        args = ['rr', 'record'] + args
     elif wrapper:
         args = wrapper.split(' ') + args
 
     p = subprocess.Popen(args, env=env)
+    print('Process', args, 'started with pid', p.pid, file=sys.stderr)
     return p.wait()
 
 
-def meta_run(klass):
+def meta_run(klass, extra_env=None, setup_argparse=None, handle_argparse=None):
     DBusGMainLoop(set_as_default=True)
 
     parser = argparse.ArgumentParser()
     parser.add_argument('--kvm', action='store_true', default=False)
     parser.add_argument('--launch', action='append', default=[])
+    parser.add_argument('--no-isolate-dirs', action='store_true', default=False)
+    if setup_argparse:
+        setup_argparse(parser)
     (args, rest) = parser.parse_known_args(sys.argv)
 
-    rest.pop(0)
-    if rest[0] == '--':
-      rest.pop(0)
+    if handle_argparse:
+        handle_argparse(args)
 
+    rest.pop(0)
+    if not rest:
+        parser.error('Command or separator `--` not found')
+    if rest[0] == '--':
+        rest.pop(0)
+    else:
+        print('WARNING: Command or separator `--` not found', file=sys.stderr)
+
+    if args.no_isolate_dirs:
+        return meta_run_klass(klass, rest,
+                              enable_kvm=args.kvm,
+                              extra_env=extra_env)
+
+    test_root = os.getenv('MUTTER_DBUS_RUNNER_TEST_ROOT')
+    if test_root:
+        print('Reusing MUTTER_DBUS_RUNNER_TEST_ROOT', test_root, file=sys.stderr)
+        return meta_run_klass(klass, rest,
+                              extra_env=extra_env)
+
+    with tempfile.TemporaryDirectory(prefix='mutter-testroot-',
+                                     ignore_cleanup_errors=True) as test_root:
+        env_dirs = [
+            'HOME',
+            'TMPDIR',
+            'XDG_CACHE_HOME',
+            'XDG_CONFIG_HOME',
+            'XDG_DATA_HOME',
+            'XDG_RUNTIME_DIR',
+        ]
+        os.environ['MUTTER_DBUS_RUNNER_TEST_ROOT'] = test_root
+        print('Setup MUTTER_DBUS_RUNNER_TEST_ROOT as', test_root, file=sys.stderr)
+        for env_dir in env_dirs:
+            directory = os.path.join(test_root, env_dir.lower())
+            os.mkdir(directory, mode=0o700)
+            os.environ[env_dir] = directory
+            print('Setup', env_dir, 'as', directory, file=sys.stderr)
+
+        return meta_run_klass(klass, rest,
+                              enable_kvm=args.kvm,
+                              launch=args.launch,
+                              bind_sockets=True,
+                              extra_env=extra_env)
+
+def meta_run_klass(klass, rest, enable_kvm=False, launch=[], bind_sockets=False, extra_env=None):
     result = 1
 
     if os.getenv('META_DBUS_RUNNER_ACTIVE') == None:
-        klass.setUpClass(args.kvm, args.launch)
+        klass.setUpClass(enable_kvm=enable_kvm,
+                         launch=launch,
+                         bind_sockets=bind_sockets)
         runner = klass()
         runner.assertGreater(len(rest), 0)
         wrapper = os.getenv('META_DBUS_RUNNER_WRAPPER')
 
         try:
             print('Running test case...', file=sys.stderr)
-            result = wrap_call(rest, wrapper)
+            result = wrap_call(rest, wrapper, extra_env)
         finally:
             MutterDBusRunner.tearDownClass()
     else:
@@ -307,7 +468,7 @@ def meta_run(klass):
                    'Not re-creating mocked environment.'),
                   file=sys.stderr)
             print('Running test case...', file=sys.stderr)
-            result = wrap_call(rest, None)
+            result = wrap_call(rest, None, extra_env)
         finally:
             pass
 
